/*
 *  HardKey Verify module. ver. 3.3
 *  with EXECryptor 2.0 support
 *  Copyright (c) 1995-2004 SoftComplete Development
 *  http://www.strongbit.com
 */

// uncomment if you using code encryption
//#define ASPROTECT
// we use EXECryptor for code encryption !
#define EXECRYPTOR

// what we use for getting hardware ID
#define HWID_HDD
//#define HWID_MAC

#include "KeyVerify.h"
#include "Key.h"
#include <stdlib.h>
#include <time.h>
#ifdef HWID_MAC
  #include <nb30.h>
#endif

#ifdef EXECRYPTOR
  #include "EXECryptor.h"
#endif

// ---------------------------- AsProtect support ------------------------------

#ifdef ASPROTECT

#include "asprotect.h"
typedef void (__stdcall *TSetKeyProc) (char   *Key, DWORD  KeySize);
TSetKeyProc SetKeyProc = NULL;

#endif

// ----------------------------------- TRC4Hash --------------------------------

BYTE RC4_State[256];
int  RC4_I, RC4_J;

void RC4_Init(const char *S, int SLen)
{ int j,i;
  BYTE K[256];
  const char *P=S;
  if (SLen < 0) SLen = strlen(S);
  for (i=0,j=SLen; i < 256; i++) {
   RC4_State[i] = i;
   K[i] = (i+13) & 0xFF;
   if (SLen > 0) { 
     K[i] = K[i] ^ (SLen + (BYTE)*P);
     if (--j == 0) {
       P=S; j=SLen;
     } else
       P++;
   }
  }
  for (RC4_I=RC4_J=j=i=0; i < 256; i++) {
    BYTE swapbyte;
    j = (j + RC4_State[i] + K[i]) & 0xFF;
    swapbyte = RC4_State[i];
    RC4_State[i] = RC4_State[j];
    RC4_State[j] = swapbyte;
  }
}

BYTE RC4_GetByte()
{ BYTE swapbyte;
  RC4_I = (RC4_I + 1) & 0xFF;
  RC4_J = (RC4_J + RC4_State[RC4_I]) & 0xFF;
  swapbyte = RC4_State[RC4_I];
  RC4_State[RC4_I] = RC4_State[RC4_J];
  RC4_State[RC4_J] = swapbyte;
  return RC4_State[(RC4_State[RC4_I] + RC4_State[RC4_J]) & 0xFF];
}

// --------------------------------- TBitStream --------------------------------

unsigned char *Int2ABC = (unsigned char *)"23456789ABCDEFGHJKLMNPQRSTUVWXYZ";
int   ABCSize = -1; //Length(Int2ABC);
int   ABC2Int[256];

void InitABC()
{ if (ABCSize < 0) {
    for (int j=0; j < 256; j++) {
      unsigned char *p=Int2ABC;
      for (; *p && *p != j; p++) ;
      if (*p == 0) ABC2Int[j]=-1;
      else ABC2Int[j]=p - Int2ABC;
    }
    ABCSize=strlen((char *)Int2ABC);
  }
}

bool IsABCChar(unsigned char ch)
{
  for (unsigned char *p=Int2ABC; *p != 0; p++)
    if (*p == ch) return true;
  return false;
}

////////////////////////////////////////////////////////////////////////////

void MsgTrim(const char *Msg, char *Dst)
{ while (*Msg) {                
    if (*Msg > ' ') *Dst++=toupper(*Msg);
    Msg++;
  }
  *Dst=0;
}

#ifdef CODECRYPT
void ExpandCodeCryptKey(DWORD CKey, unsigned char *EAPKey)
{ unsigned char TKey[256];
  itoa(CKey,TKey,10);
  RC4_Init(TKey,-1);
  for (int i=0; i < 1024; i++) {
    for (int j=0; j < 256; j++)
      TKey[j] = RC4_GetByte();
    RC4_Init(TKey,256);
  }
  for (int j=0; j < 60; j++)
    *EAPKey++ = RC4_GetByte();
}

bool Called = false;
#endif

void HFE(unsigned char *MsgBits, unsigned char *SignBits)
{
  unsigned char S[SignSize],M,mask,bmask,*p;
  int j,i,k;
#define GetBit() \
  (*p & bmask ? 1 : 0)
#define NextBit() \
  { bmask <<= 1; \
    if (bmask == 0) { \
      p++; bmask=1; \
    } \
  }
  mask=1;
  for (j=0; j < SignSize; j++) {
    S[j] = (*SignBits & mask ? 1 : 0);
    mask <<= 1;
    if (mask == 0) {
      mask=1;
      SignBits++;
    }
  }
  p=PK;
  mask=bmask=1;
  for (j=0; j < MsgSize; j++) {
    M=GetBit();
    NextBit();
    for (i=0; i < SignSize; i++) {
      if (GetBit()) M+=S[i];
      NextBit();
    }
    for (i=0; i < SignSize-1; i++)
      for (k=i+1; k < SignSize; k++) {
        if (GetBit()) M+=S[i]*S[k];
        NextBit();
      }
    if (M & 1) *MsgBits |= mask;
    else *MsgBits &= ~mask;
    mask <<= 1;
    if (mask == 0) {
      mask=1;
      MsgBits++;
    }
    if (bmask > 1) {
      p++;
      bmask = 1;
    }
  }
}

bool VerifyRegCode(
  const char *RegName,
  const char *RegCode,
  short int *LicType,
  char *LicName,
  char *LicInfo,
  int *Expire) // // days from 1 Jan 2002
{
  Int64 hMsg=0;
  char RegNameTrimmed[256];
  unsigned char MsgHash[256], MsgBits[16], SignBits[16], *p, mask;
#define WriteBits(v,_cnt) \
  { DWORD vmask=1; \
    int cnt=_cnt; \
    while (cnt > 0) { \
      if (v & vmask) *p |= mask; \
      else *p &= ~mask; \
      vmask <<= 1; cnt--; mask <<= 1; \
      if (mask == 0) { \
        mask=1; p++; \
      } \
    } \
  }
#define ReadBits(_res,_cnt) \
  { DWORD vmask=1; \
    int cnt=_cnt; \
    _res=0; \
    while (cnt > 0) { \
      if (*p & mask) _res|=vmask; \
      vmask <<= 1; cnt--; mask <<= 1; \
      if (mask == 0) { \
        mask=1; p++; \
      } \
    } \
  }
#ifdef EXECRYPTOR
  CRYPT_START
#endif
  InitABC();
  *LicType=0; *LicName=*LicInfo='\0';
  *Expire=0;
  //
  MsgTrim(RegName,RegNameTrimmed);
  RC4_Init(RegNameTrimmed,-1);
#ifdef EXECRYPTOR
  CRYPT_END
#endif
  for (int i=0; i < 256; i++) {
    for (int j=0; j < 256; j++)
      MsgHash[j]=RC4_GetByte();
    RC4_Init((char *)MsgHash,256);
  }
  for (int j=0; j < 7; j++)
    MsgHash[j]=RC4_GetByte();
  for (int j=0; j < 8; j++) {
    hMsg <<= 8;
    hMsg |= RC4_GetByte();
  }
#ifdef EXECRYPTOR
  CRYPT_START_UNSAFE
#ifdef CODECRYPT
  EXECryptor_SetCodeKey(NULL, 0);
#endif
#endif
  for (int j=0; j < BlackListCount; j++)
    if (BlackList[j] == hMsg) {
      return false;
    }
  p=SignBits;
  mask=1;
  for (; *RegCode; RegCode++)
    if (IsABCChar(*RegCode))
      WriteBits(ABC2Int[(unsigned char)*RegCode],5);
  HFE(MsgBits, SignBits);
  for (int j=0; j < 5; j++)
    if (MsgBits[j] != MsgHash[j]) {
      return false;
    }
  if ((MsgBits[6] & 3) != (MsgHash[6] & 3))
    return false;
  p=MsgBits+6;
  mask=0x20;
  ReadBits(*Expire,12);
  if (*Expire > 0) {
    time_t CurTime;
    struct tm ETime;
    time(&CurTime);
    CurTime /= 24*60*60;
    ETime.tm_year = 2002 - 1900;
    ETime.tm_mon  = 0;
    ETime.tm_mday = 1;
    ETime.tm_hour = 0;
    ETime.tm_min  = 0;
    ETime.tm_sec  = 0;
    ETime.tm_isdst = 0;
    if (*Expire + mktime(&ETime)/(24*60*60) + 1 <= CurTime)
      return false;
  }
  p=MsgBits+8;
  mask=2;
  #ifdef EXECRYPTOR
    CRYPT_END_UNSAFE
  #endif
  #ifdef CODECRYPT
    #ifdef ASPROTECT
      if (! Called) {
        if (SetKeyProc != NULL) {
          DWORD CCKey;
          unsigned char ExpandedKey[60];
          ReadBits(CCKey,25);
          ExpandCodeCryptKey(CCKey,ExpandedKey);
          (*SetKeyProc)(ExpandedKey, 60);
        }
        Called=true;
      }
      REG_CRYPT_BEGIN
    #endif
    #ifdef EXECRYPTOR
      {
        DWORD CCKey;
        unsigned char ExpandedKey[60];
        ReadBits(CCKey,25);
        ExpandCodeCryptKey(CCKey,ExpandedKey);
        EXECryptor_SetCodeKey(ExpandedKey, 60);
      }
      CRYPT_REG_UNSAFE
    #endif
    p=MsgBits+11;
    mask=4;
  #endif
  #ifdef USEHWID
    { DWORD HWID;
      ReadBits(HWID,25);
      if ((HWIDProc != NULL) && (HWID != 0))
        if (HWID != ((*HWIDProc)() & 0x1FFFFFF))
          return false;
    }
  #endif
  p=MsgBits+6;
  mask=4;
  ReadBits(*LicType,3);
  if (*LicType >= 0 && *LicType < LicCount) {
    strcpy(LicName,LicListName[*LicType]);
    strcpy(LicInfo,LicListInfo[*LicType]);
  }
  #ifdef CODECRYPT
  #ifdef EXECRYPTOR
    return true;
    CRYPT_END_UNSAFE
  #endif
  #ifdef ASPROTECT
    REG_CRYPT_END
    return true;
  #endif
  #else
    return true;
  #endif
}

#ifndef EXECRYPTOR
DWORD GetDiskSerialNo()
{ DWORD mc,fl,DiskSerNo;
  char buff[256];
  GetWindowsDirectory(buff,255);
  if (*buff == 0) {
    GetVolumeInformation("C:\\",NULL,0,&DiskSerNo,&mc,&fl,NULL,0);
  } else {
    buff[1]=':';
    buff[2]='\\';
    buff[3]=0;
    GetVolumeInformation(buff,NULL,0,&DiskSerNo,&mc,&fl,NULL,0);
  }
  return DiskSerNo;
}
#endif

// Get the list of adapters
unsigned char GetLanaEnum(PLANA_ENUM LanaEnum)
{ NCB LanaEnumNCB;
  FillMemory(&LanaEnumNCB, sizeof(LanaEnumNCB), 0);
  LanaEnumNCB.ncb_buffer = (PUCHAR)LanaEnum;
  LanaEnumNCB.ncb_length = sizeof(LANA_ENUM);
  LanaEnumNCB.ncb_command = (UCHAR)NCBENUM;
  Netbios(&LanaEnumNCB);
  return LanaEnumNCB.ncb_cmd_cplt;
}

#define MACAddrSize 6

unsigned char GetMACAddress(unsigned char LanaNum, unsigned char *MACAddress)
{ unsigned char Result;
  PADAPTER_STATUS AdapterStatus;
  NCB StatNCB;
  FillMemory(&StatNCB, sizeof(StatNCB), 0);
  StatNCB.ncb_lana_num = LanaNum;
  StatNCB.ncb_command = NCBRESET;
  Netbios(&StatNCB);
  FillMemory(&StatNCB, sizeof(StatNCB), 0);
  StatNCB.ncb_length = sizeof(ADAPTER_STATUS) +  255 * sizeof(NAME_BUFFER);
  AdapterStatus = (PADAPTER_STATUS)malloc(StatNCB.ncb_length);
  FillMemory(MACAddress, MACAddrSize*sizeof(unsigned char), 0);
  StatNCB.ncb_buffer = (PUCHAR)AdapterStatus;
  MoveMemory(StatNCB.ncb_callname,"*              ",16);
  StatNCB.ncb_lana_num = LanaNum;
  StatNCB.ncb_command = NCBASTAT;
  Netbios(&StatNCB);
  Result = StatNCB.ncb_cmd_cplt;
  if (Result == NRC_GOODRET)
    MoveMemory(MACAddress, AdapterStatus, MACAddrSize*sizeof(unsigned char));
  free(AdapterStatus);
  return Result;
}

DWORD GetMACAddr()
{ LANA_ENUM LanaEnum;
  unsigned char MACAddress[MACAddrSize];
  DWORD Result;
  Result=0;
  if (GetLanaEnum(&LanaEnum) == NRC_GOODRET)
    if ((unsigned char)LanaEnum.length > 0)
      if (GetMACAddress((unsigned char)LanaEnum.lana[0], MACAddress) == NRC_GOODRET)
        Result = *(DWORD*)MACAddress ^ *(DWORD*)(MACAddress+2);
  return Result;
}

static HWID_Func HWIDProc = GetHWID_Std;

DWORD GetHWID_Std()
{
  DWORD Result=0;
  #ifdef HWID_HDD
    #ifdef EXECRYPTOR
      Result ^= EXECryptor_GetHardwareID();
    #else
      Result ^= GetDiskSerialNo();
    #endif
  #endif
  #ifdef HWID_MAC
    Result ^= GetMACAddr();
  #endif
  return Result;
}

static char HWID_STR[6] = "";
const char* GetHardwareID()
{ if (*HWID_STR == 0) {
#ifdef EXECRYPTOR
    CRYPT_START
#endif
  if (HWIDProc != NULL) {
    DWORD HWID=(*HWIDProc)();
    InitABC();
    for (int j=0; j < 5; j++) {
      HWID_STR[j]=Int2ABC[HWID & 0x1F];
      HWID >>= 5;
    }
    HWID_STR[5]='\0';
  }
#ifdef EXECRYPTOR
    CRYPT_END
#endif
  }
  return HWID_STR;
}

void EncryptDecryptStr(const char *Src, int SrcLen, char *Dst)
{
  RC4_Init(GetHardwareID(),-1);
  for (int k=0; (SrcLen < 0 && *Src) || (SrcLen >= 0 && k < SrcLen); k++)
    *Dst++ = *Src++ ^ RC4_GetByte() ^ PK[k];
  *Dst = '\0';
}

// hexadecimal digits/letter
static char i2h[] = "0123456789ABCDEF";

#ifndef EXECRYPTOR
static char SecureGUID[128] = "";
const char *GetSecureGUID()
{ // generate fake GUID (depend on hardware id and application public key)
  if (*SecureGUID == 0) {
    strcat(SecureGUID,"CLS");
    strcat(SecureGUID,"ID\\{");
    RC4_Init(GetHardwareID(),-1);
    int k=0;
    char *p=SecureGUID+strlen(SecureGUID);
    for (int j=0; j < 8; j++)
      *p++ = i2h[(RC4_GetByte() ^ PK[k++]) & 0xF];
    *p++ = '-';
    for (int i=0; i < 3; i++) {
      for (int j=0; j < 4; j++)
        *p++ = i2h[(RC4_GetByte() ^ PK[k++]) & 0xF];
      *p++ = '-';
    }
    for (int j=0; j < 12; j++)
      *p++ = i2h[(RC4_GetByte() ^ PK[k++]) & 0xF];
    *p = '\0';
    strcat(SecureGUID,"}\\In");
    strcat(SecureGUID,"Proc");
    strcat(SecureGUID,"Serv");
    strcat(SecureGUID,"er32");
  }
  return SecureGUID;
}

void HashName(const char *Name, char *Hashed)
{
  RC4_Init(GetHardwareID(),-1);
  for (int k=0; *Name; k++)
    *Hashed++ = 'a' + (*Name++ ^ RC4_GetByte() ^ PK[k]) % ('z' - 'a');
  *Hashed = '\0';
}
#endif

const char*
  UnsafeRegKey = "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\";

void SecureRead(const char *Name, char *Value)
{
#ifdef EXECRYPTOR
  EXECryptor_SecureRead(Name,Value);
#else
  HKEY KeyHandle;
  char Name2[256];
  DWORD sz,tp;
  *Value = '\0';
  strcpy(Name2,UnsafeRegKey);
  strcat(Name2,GetSecureGUID());
  if ((RegOpenKeyExA(HKEY_CLASSES_ROOT,
       GetSecureGUID(), 0, KEY_QUERY_VALUE, &KeyHandle) == ERROR_SUCCESS) ||
      (RegOpenKeyExA(HKEY_CURRENT_USER,
       Name2, 0, KEY_QUERY_VALUE, &KeyHandle) == ERROR_SUCCESS)) {
    HashName(Name,Name2);
    tp=REG_BINARY;
    if (RegQueryValueExA(KeyHandle, Name2, NULL, &tp, NULL, &sz) == ERROR_SUCCESS) {
      RegQueryValueExA(KeyHandle, Name2, NULL, &tp, Value, &sz);
      EncryptDecryptStr(Value,sz,Value);
    }
    RegCloseKey(KeyHandle);
  }
#endif
}

void SecureWrite(const char *Name, const char *Value)
{
#ifdef EXECRYPTOR
  EXECryptor_SecureWrite(Name,Value);
#else
  HKEY KeyHandle;
  DWORD Disposition;
  char Name2[256],*Value2;
  strcpy(Name2,UnsafeRegKey);
  strcat(Name2,GetSecureGUID());
  if ((RegCreateKeyExA(HKEY_CLASSES_ROOT, GetSecureGUID(), 0, NULL,
       REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, NULL, &KeyHandle, &Disposition) == ERROR_SUCCESS) ||
      (RegCreateKeyExA(HKEY_CURRENT_USER, Name2, 0, NULL,
       REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, NULL, &KeyHandle, &Disposition) == ERROR_SUCCESS)) {
    HashName(Name,Name2);
    Value2 = new char[strlen(Value)+1];
    EncryptDecryptStr(Value,-1,Value2);
    RegSetValueExA(KeyHandle, Name2, 0, REG_BINARY, Value2, strlen(Value));
    RegCloseKey(KeyHandle);
    delete Value2;
  }
#endif
}

//------------------------------------------------------------------------------
#ifdef ASPROTECT
void __declspec(dllexport) __stdcall SetDecryptionKey(TSetKeyProc DProc)
{
  SetKeyProc = DProc;
}
#endif
