unit KeyVerify;

{   HardKey Verify module. ver. 3.3
    with EXECryptor 2.0 support
    Copyright (c) 1995-2004 SoftComplete Development
    http://www.strongbit.com
}

interface

// remove dot if you using code encryption
{-$DEFINE ASPROTECT}
{$DEFINE EXECRYPTOR} // we use EXECryptor code encryption

uses Windows, SysUtils, Classes;

type
  TLicType = 0..7;
  THWIDProc = function: DWORD;

function VerifyRegCode(const RegName, RegCode: string;
  var LicType: TLicType; var LicName,LicInfo: string;
  var Expire: TDateTime): Boolean;

function GetHardwareID: string;

// secure registry storage
function SecureRead(const Name: string): string;
procedure SecureWrite(const Name,Value: string);

// current date
function SecureGetDate: TDateTime;

// support for user defined Hardware ID
function GetHWID_Std: DWORD;
var HWIDProc: THWIDProc = GetHWID_Std; // set this for customize hardware ID

// what we use for getting hardware ID
{$DEFINE HWID_HDD}
{.$DEFINE HWID_MAC}

implementation

{$IFDEF EXECRYPTOR}
  {$IFDEF HWID_MAC}
  uses NB30, EXECryptor;
  {$ELSE}
  uses EXECryptor;
  {$ENDIF}
{$ELSE}
  {$IFDEF HWID_MAC}
  uses NB30;
  {$ENDIF}
{$ENDIF}

{$IFDEF EXECRYPTOR}
{$DEFINE UNSAFE_CRYPT}
{$ENDIF}

// ---------------------------- AsProtect support ------------------------------

{$IFDEF ASPROTECT}
type
  TSetKeyProc = procedure(Key : Pointer;KeySize : DWORD);stdcall;
  TProcedure  = procedure;

const
  SetKeyProc    : TSetKeyProc = nil;
{$ENDIF}

// -----------------------------------------------------------------------------

type
  TRC4Hash = class
    FState: Array[0..255] of BYTE;
    FI,FJ: BYTE;
    procedure Init(const S: string);
    function GetByte: byte;
  end;

const
  Int2ABC = '23456789ABCDEFGHJKLMNPQRSTUVWXYZ';
  ABCSize = Length(Int2ABC);
var
  ABC2Int: array[Char] of integer;

function IsABCChar(ch: char): Boolean;
begin
  Result:=Pos(ch,Int2ABC) > 0;
end;

{$I key.inc}

function VerifyRegCode(const RegName, RegCode: string;
  var LicType: TLicType; var LicName,LicInfo: string;
  var Expire: TDateTime): Boolean;

const
  Called: Boolean = False;

  function MsgTrim(const Msg: string): string;
  var j: integer;
  begin
    Result:='';
    for j:=1 to Length(Msg) do
      if Ord(Msg[j]) > Ord(' ') then
        Result:=Result+UpperCase(Msg[j]);
  end;

  procedure HFE(MsgBits, SignBits: PBYTE);
  var S: array of byte;
      M: byte;
      mask,bmask: byte;
      p: PBYTE;
      j,i,k: integer;
    function GetBit: byte;
    begin
      if (p^ and bmask) <> 0 then Result:=1
      else Result:=0;
      bmask:=bmask shl 1;
      if bmask = 0 then begin
        Inc(p);
        bmask:=1;
      end;
    end;
  begin
    SetLength(S,SignSize);
    mask:=1;
    for j:=0 to SignSize-1 do begin
      if (SignBits^ and mask) <> 0 then S[j]:=1
      else S[j]:=0;
      mask := mask shl 1;
      if mask = 0 then begin
        mask:=1;
        Inc(SignBits);
      end;
    end;
    p:=@PK[0];
    mask:=1;
    bmask:=1;
    for j:=0 to MsgSize-1 do begin
      M:=GetBit;
      for i:=0 to SignSize-1 do
        if GetBit <> 0 then Inc(M,S[i]);
      for i:=0 to SignSize-2 do
        for k:=i+1 to SignSize-1 do
          if GetBit <> 0 then Inc(M,S[i]*S[k]);
      if (M and 1) <> 0 then MsgBits^:=MsgBits^ or mask
      else MsgBits^:=MsgBits^ and not mask;
      mask:=mask shl 1;
      if mask = 0 then begin
        mask:=1;
        Inc(MsgBits);
      end;
      if bmask > 1 then begin
        Inc(p);
        bmask := 1;
      end;
    end;
  end;

{$IFDEF CODECRYPT}
  function ExpandCodeCryptKey(CKey: DWORD): string;
  var i,j: integer;
  begin
    with TRC4Hash.Create do try
      Init(IntToStr(CKey));
      SetLength(Result,256);
      for i:=1 to 1024 do begin
        for j:=1 to 256 do
          Result[j]:=Chr(GetByte);
        Init(Result);
      end;
      SetLength(Result,60);
      for j:=1 to 60 do
        Result[j]:=Chr(GetByte);
    finally
      Free;
    end;
  end;
{$ENDIF}

var
  MsgBits, SignBits: array[0..15] of byte;
  mask: byte;
  p: PBYTE;
  j,i: integer;
  days: integer;
  HWID: DWORD;
  MsgHash {$IFDEF CODECRYPT}, ExpandedKey {$ENDIF} : string;
  hMsg: int64;

  procedure WriteBits(v: DWORD; cnt: integer);
  var vmask: DWORD;
  begin
    vmask:=1;
    while cnt > 0 do begin
      if (v and vmask) <> 0 then p^:=p^ or mask
      else p^:=p^ and not mask;
      vmask:=vmask shl 1;
      Dec(cnt);
      mask:=mask shl 1;
      if mask = 0 then begin
        mask:=1;
        Inc(p);
      end;
    end;
  end;

  function ReadBits(cnt: integer): DWORD;
  var vmask: DWORD;
  begin
    Result:=0;
    vmask:=1;
    while cnt > 0 do begin
      if (p^ and mask) <> 0 then Result:=Result or vmask;
      vmask:=vmask shl 1;
      Dec(cnt);
      mask:=mask shl 1;
      if mask = 0 then begin
        mask:=1;
        Inc(p);
      end;
    end;
  end;
  
begin
{$IFDEF EXECRYPTOR}
{$I crypt_start.inc}
{$ENDIF}
  LicType:=0; LicName:=''; LicInfo:='';
  Expire:=0;
  Result:=True;
{$IFDEF EXECRYPTOR}
{$I crypt_end.inc}
{$ENDIF}
  with TRC4Hash.Create do try
    Init(MsgTrim(RegName));
    SetLength(MsgHash,256);
    for i:=1 to 256 do begin
      for j:=1 to 256 do
        MsgHash[j]:=Chr(GetByte);
      Init(MsgHash);
    end;
    {$IFDEF EXECRYPTOR}
    {$I crypt_start.inc}
    {$ENDIF}
    SetLength(MsgHash,7);
    for j:=1 to 7 do
      MsgHash[j]:=Chr(GetByte);
    hMsg:=0;
    for j:=1 to 8 do
      hMsg:=(hMsg shl 8) or GetByte;
    {$IFDEF EXECRYPTOR}
    {$I crypt_end.inc}
    {$ENDIF}
  finally
    Free;
  end;
{$IFDEF EXECRYPTOR}
{$I crypt_start.inc}
{$IFDEF CODECRYPT}
  SetCodeKey(nil, 0);
{$ENDIF}
{$ENDIF}
  if BlackListCount > 0 then
    for j:=0 to BlackListCount-1 do
      if BlackList[j] = hMsg then begin
        Result:=False;
        Exit;
      end;
  if Result then begin
    p:=@SignBits[0];
    mask:=$1;
    for j:=1 to Length(RegCode) do
      if IsABCChar(RegCode[j]) then
        WriteBits(ABC2Int[RegCode[j]],5);
    HFE(@MsgBits[0], @SignBits[0]);
    for j:=0 to 5 do
      if MsgBits[j] <> Ord(MsgHash[j+1]) then begin
        Result:=False;
        Exit;
      end;
    if (MsgBits[6] and $3) <> (Ord(MsgHash[7]) and $3) then begin
      Result:=False;
      Exit;
    end;
    if Result then begin
      p:=@MsgBits[6];
      mask:=$20;
      days:=ReadBits(12);
      if days > 0 then begin
        Expire:=EncodeDate(2002,1,1)+days;
        if Expire < SecureGetDate then begin
          Result:=False;
          Exit;
        end;
      end else
        Expire:=0;
      {$IFDEF EXECRYPTOR}
      {$I crypt_end.inc}
      {$ENDIF}
      if Result then begin
        p:=@MsgBits[8];
        mask:=$2;
        {$IFDEF CODECRYPT}
          {$IFDEF ASPROTECT}
            if not Called then begin
              if Assigned(SetKeyProc) then begin
                ExpandedKey := ExpandCodeCryptKey(ReadBits(25));
                SetKeyProc(PChar(ExpandedKey), Length(ExpandedKey));
              end;
              Called:=True;
            end;
            {$I reg_crypt_begin.inc}
          {$ENDIF}
          {$IFDEF EXECRYPTOR}
            ExpandedKey := ExpandCodeCryptKey(ReadBits(25));
            SetCodeKey(PChar(ExpandedKey),Length(ExpandedKey));
            {$I crypt_reg.inc}
          {$ENDIF}
          p:=@MsgBits[11];
          mask:=$4;
        {$ENDIF}
        {$IFDEF USEHWID}
          HWID:=ReadBits(25);
          if Assigned(HWIDProc) and (HWID <> 0) then
            if HWID <> (HWIDProc and $1FFFFFF) then begin
              Result:=False;
              {$IFNDEF ASPROTECT}
              Exit;
              {$ENDIF}
            end;
        {$ENDIF}
        p:=@MsgBits[6];
        mask:=$4;
        if Result then begin
          LicType:=ReadBits(3);
          if (LicType >= Low(LicList)) and (LicType <= High(LicList)) then begin
            LicName:=LicList[LicType,0];
            LicInfo:=LicList[LicType,1];
          end;
        end;
        {$IFDEF CODECRYPT}
        {$IFDEF EXECRYPTOR}
        {$I crypt_end.inc}
        {$ENDIF}
        {$IFDEF ASPROTECT}
        {$I reg_crypt_end.inc}
        {$ENDIF}
        {$ENDIF}
      end;
    end;
  end;
end;

{ TRC4Hash }

procedure TRC4Hash.Init(const S: string);
var j: integer;
    i, swapbyte: BYTE;
    K: array[0..255] of BYTE;
begin
  FI := 0; FJ := 0; j := 0;
  for i:= 0 to 255 do begin
   FState[i] := i;
   K[i] := (i+13) and $FF;
   if Length(S) > 0 then
     K[i] := K[i] xor (Length(S) + BYTE(S[(i MOD Length(S))+1]));
  end;
  for i := 0 to 255 do begin
    j := (j + FState[i] + K[i]) MOD 256;
    swapbyte := FState[i];
    FState[i] := FState[j];
    FState[j] := swapbyte;
  end;
end;

function TRC4Hash.GetByte: byte;
var swapbyte: BYTE;
begin
  FI := (FI + 1) MOD 256;
  FJ := (FJ + FState[FI]) MOD 256;
  swapbyte := FState[FI];
  FState[FI] := FState[FJ];
  FState[FJ] := swapbyte;
  Result := FState[(FState[FI] + FState[FJ]) MOD 256];
end;

procedure InitABC;
var j: char;
begin
  for j:= #0 to #255 do
    ABC2Int[j]:=Pos(j,Int2ABC)-1;
end;

function GetHWID_Std: DWORD;
{$IFDEF HWID_MAC}
  type
   TMACAddress = array[0..5] of Byte;
   PMACAddress = ^TMACAddress;
{$ENDIF}

{$IFDEF HWID_HDD}
{$IFNDEF EXECRYPTOR}
  function GetDiskSerialNo: DWORD;
  var
    mc,fl : dword;
    buff: string;
  begin
    SetLength(buff,1024);
    SetLength(buff,GetWindowsDirectoryA(PChar(buff),1024));
    if buff = '' then
      GetVolumeInformationA(PChar('C'+':'+'\'),nil,0,@Result,mc,fl,nil,0)
    else begin
      buff:=buff[1]+':\';
      GetVolumeInformationA(PChar(buff),nil,0,@Result,mc,fl,nil,0);
    end;
  end;
{$ENDIF}
{$ENDIF}

{$IFDEF HWID_MAC}
  function GetMACAddr: DWORD;
  var
    LanaEnum: PLanaEnum;
    MACAddress: PMACAddress;

    // Get the list of adapters
    function GetLanaEnum(LanaEnum: PLanaEnum): Byte;
    var
      LanaEnumNCB: PNCB;
    begin
      New(LanaEnumNCB);
      ZeroMemory(LanaEnumNCB, SizeOf(TNCB));
      try
        with LanaEnumNCB^ do
        begin
          ncb_buffer := PChar(LanaEnum);
          ncb_length := SizeOf(TLanaEnum);
          ncb_command  := Char(NCBENUM);
          NetBios(LanaEnumNCB);
          Result := Byte(ncb_cmd_cplt);
        end;
      finally
        Dispose(LanaEnumNCB);
      end;
    end;

    function GetMACAddress(LanaNum: Byte; MACAddress: PMACAddress): Byte;
    var
      AdapterStatus: PAdapterStatus;
      StatNCB: PNCB;
    begin
      New(StatNCB);
      ZeroMemory(StatNCB, SizeOf(TNCB));
      StatNCB.ncb_length := SizeOf(TAdapterStatus) +  255 * SizeOf(TNameBuffer);
      GetMem(AdapterStatus, StatNCB.ncb_length);
      try
        with StatNCB^ do begin
          // reset NCB
          ZeroMemory(MACAddress, SizeOf(TMACAddress));
          ncb_lana_num := Char(LanaNum);
          ncb_command  := Char(NCBRESET);
          NetBios(StatNCB);
          // fetch mac address
          ZeroMemory(MACAddress, SizeOf(TMACAddress));
          ncb_buffer := PChar(AdapterStatus);
          ncb_callname := '*              ' + #0;
          ncb_lana_num := Char(LanaNum);
          ncb_command  := Char(NCBASTAT);
          NetBios(StatNCB);
          Result := Byte(ncb_cmd_cplt);
          if Result = NRC_GOODRET then
            MoveMemory(MACAddress, AdapterStatus, SizeOf(TMACAddress));
        end;
      finally
        FreeMem(AdapterStatus);
        Dispose(StatNCB);
      end;
    end;

  begin
    Result:=0;
    New(LanaEnum);
    ZeroMemory(LanaEnum, SizeOf(TLanaEnum));
    try
      if GetLanaEnum(LanaEnum) = NRC_GOODRET then
        if Byte(LanaEnum.length) > 0 then begin
          New(MACAddress);
          try
            if GetMACAddress(Byte(LanaEnum.lana[0]), MACAddress) = NRC_GOODRET then
              Result := PDWORD(MACAddress)^ xor PDWORD(@MACAddress[2])^;
          finally
            Dispose(MACAddress);
          end;
        end;
    finally
      Dispose(LanaEnum);
    end;
  end;
{$ENDIF}

begin
  Result:=0;
  {$IFDEF HWID_HDD}
  {$IFDEF EXECRYPTOR}
    Result:=Result xor EXECryptor.GetHardwareID;
  {$ELSE}
    Result:=Result xor GetDiskSerialNo;
  {$ENDIF}
  {$ENDIF}
  {$IFDEF HWID_MAC}
    Result:=Result xor GetMACAddr;
  {$ENDIF}
end;

function GetHardwareID: string;
var j: integer;
    HWID: DWORD;
begin
  {$IFDEF EXECRYPTOR}
    {$I crypt_start.inc}
  {$ENDIF}
  if Assigned(HWIDProc) then begin
    HWID:=HWIDProc;
    SetLength(Result,5);
    for j:=1 to 5 do begin
      Result[j]:=Int2ABC[(HWID and $1F)+1];
      HWID:=HWID shr 5;
    end;
  end else
    Result:='';
  {$IFDEF EXECRYPTOR}
    {$I crypt_end.inc}
  {$ENDIF}
end;

function EncryptDecryptStr(const S: string; SLen: integer = -1): string;
var j,k: integer;
begin
  with TRC4Hash.Create do try
    Init(GetHardwareID);
    k:=0;
    if SLen < 0 then SLen:=Length(S);
    SetLength(Result,SLen);
    for j:=1 to SLen do begin
      Result[j]:=Chr(Ord(S[j]) xor GetByte xor PK[k]);
      Inc(k);
    end;
  finally
    Free;
  end;
end;

{$IFNDEF EXECRYPTOR}
function GetSecureGUID: string;
var j,i,k: integer;
begin
  // generate fake GUID (depend on hardware id and application public key) 
  Result:='';
  with TRC4Hash.Create do try
    Init(GetHardwareID);
    k:=0;
    for j:=1 to 8 do begin
      Result:=Result+IntToHex((GetByte xor PK[k]) and $F,1);
      Inc(k);
    end;
    Result:=Result+'-';
    for i:=1 to 3 do begin
      for j:=1 to 4 do begin
        Result:=Result+IntToHex((GetByte xor PK[k]) and $F,1);
        Inc(k);
      end;
      Result:=Result+'-';
    end;
    for j:=1 to 12 do begin
      Result:=Result+IntToHex((GetByte xor PK[k]) and $F,1);
      Inc(k);
    end;
  finally
    Free;
  end;
  Result:=PChar('CLS')+'ID\{'+Result+'}\In'+'Proc'+'Serv'+'er32';
end;

function HashName(const Name: string): string;
var j,k: integer;
begin
  with TRC4Hash.Create do try
    Init(GetHardwareID);
    k:=0;
    SetLength(Result,Length(Name));
    for j:=1 to Length(Name) do begin
      Result[j]:=Chr(Ord('a') + (Ord(Name[j]) xor GetByte xor PK[k]) mod (Ord('z')-Ord('a')));
      Inc(k);
    end;
  finally
    Free;
  end;
end;
{$ENDIF}

const
  UnsafeRegKey = 'Software\Microsoft\Windows\CurrentVersion\Explorer\';

function SecureRead(const Name: string): string;
{$IFNDEF EXECRYPTOR}
var KeyHandle: HKEY;
    KeyGUID,Name2: string;
    sz,tp: DWORD;
{$ENDIF}
begin
{$IFDEF EXECRYPTOR}
  EXECryptor.SecureRead(Name,Result);
{$ELSE}
  KeyGUID:=GetSecureGUID;
  if RegOpenKeyExA(HKEY_CLASSES_ROOT, PChar(KeyGUID), 0,
       KEY_QUERY_VALUE, KeyHandle) <> ERROR_SUCCESS then
    if RegOpenKeyExA(HKEY_CURRENT_USER, PChar(UnsafeRegKey+KeyGUID), 0,
         KEY_QUERY_VALUE, KeyHandle) <> ERROR_SUCCESS then begin
      Result:='';
      Exit;
    end;
  try
    Name2:=HashName(Name);
    tp:=REG_BINARY;
    if RegQueryValueExA(KeyHandle, PChar(Name2), nil, @tp, nil, @sz) = ERROR_SUCCESS then begin
      SetLength(Result,sz);
      RegQueryValueExA(KeyHandle, PChar(Name2), nil, @tp, PByte(PChar(Result)), @sz);
      Result:=EncryptDecryptStr(Result,sz);
    end;
  finally
    RegCloseKey(KeyHandle);
  end;
{$ENDIF}
end;

procedure SecureWrite(const Name,Value: string);
{$IFNDEF EXECRYPTOR}
var KeyHandle: HKEY;
    KeyGUID,Value2,Name2: string;
    Disposition: integer;
{$ENDIF}
begin
{$IFDEF EXECRYPTOR}
  EXECryptor.SecureWrite(Name,Value);
{$ELSE}
  KeyGUID:=GetSecureGUID;
  if RegCreateKeyExA(HKEY_CLASSES_ROOT, PChar(KeyGUID), 0, nil,
      REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, nil, KeyHandle, @Disposition) = ERROR_SUCCESS then
    if RegCreateKeyExA(HKEY_CURRENT_USER, PChar(UnsafeRegKey+KeyGUID), 0, nil,
        REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, nil, KeyHandle, @Disposition) = ERROR_SUCCESS then
      Exit;
  try
    Name2:=HashName(Name);
    Value2:=EncryptDecryptStr(Value);
    RegSetValueExA(KeyHandle, PChar(Name2), 0, REG_BINARY, PChar(Value2),  Length(Value));
  finally
    RegCloseKey(KeyHandle);
  end;
{$ENDIF}
end;

function SecureGetDate: TDateTime;
{$IFNDEF EXECRYPTOR}
var
  SystemTime: TSystemTime;
{$ENDIF}
begin
{$IFDEF EXECRYPTOR}
  Result:=EXECryptor.SafeGetDate;
{$ELSE}
  GetLocalTime(SystemTime);
  with SystemTime do Result := EncodeDate(wYear, wMonth, wDay);
{$ENDIF}
end;

//----------------------------------------------------------------------
// This function gets pointer on function for setting of external key
//------------------------------------------------------------------------------
{$IFDEF ASPROTECT}
procedure SetDecryptionKey(DProc : TSetKeyProc);stdcall;
begin
  SetKeyProc := DProc;
end;

exports
  SetDecryptionKey;
{$ENDIF}

initialization
  InitABC;
  Assert(ABCSize = 32);
end.


